import 'dart:math' as math;

import 'context.dart';
import 'env.dart';
import 'runtime.dart';
import 'tests.dart';
import 'utils.dart';

abstract class Node {
  const Node();

  String render(Context context) => '';

  @override
  String toString() => 'Node()';
}

// Nodes

class Output extends Node {
  static Node orNode(List<Node> nodes) {
    if (nodes.isEmpty) return const Literal<String>('');
    return nodes.length > 1 ? Output(nodes) : nodes.first;
  }

  const Output(this.nodes);

  final List<Node> nodes;

  @override
  String render(Context context) =>
      nodes.map((node) => node.render(context)).join();

  @override
  bool operator ==(dynamic other) => other is Output && other.nodes == nodes;

  @override
  int get hashCode => nodes.hashCode ^ 439390411;

  @override
  String toString() => 'Output($nodes)';
}

abstract class Expression extends Node {
  const Expression();

  dynamic call(Context context);

  @override
  String render(Context context) => '${context.env.finalize(call(context))}';
}

class ExpressionList<T> extends Expression {
  const ExpressionList(this.values);
  ExpressionList.of(Iterable<Expression> values)
      : this.values = List<Expression>.unmodifiable(values);

  final List<Expression> values;

  List<T> call(Context context) =>
      values.map<T>((value) => value(context) as T).toList();

  @override
  bool operator ==(dynamic other) =>
      other is ExpressionList<T> && other.values == values;

  @override
  int get hashCode => T.hashCode ^ values.hashCode ^ 571751238;

  @override
  String toString() => 'ExpressionList<$T>($values)';
}

class ExpressionMap<T> extends Expression {
  const ExpressionMap(this.values);
  ExpressionMap.of(Map<String, Expression> values)
      : this.values = Map<String, Expression>.unmodifiable(values);

  final Map<String, Expression> values;

  Map<String, T> call(Context context) => values.map<String, T>(
      (key, value) => MapEntry<String, T>(key, value(context) as T));

  @override
  bool operator ==(dynamic other) =>
      other is ExpressionMap<T> && other.values == values;

  @override
  int get hashCode => T.hashCode ^ values.hashCode ^ 738574191;

  @override
  String toString() => 'ExpressionMap<$T>($values)';
}

class Literal<T> extends Expression {
  const Literal(this.value);

  final T value;

  T call(_) => value;

  @override
  String render(_) => '$value';

  @override
  bool operator ==(dynamic other) =>
      other is Literal<T> && other.value == value;

  @override
  int get hashCode => T.hashCode ^ value.hashCode ^ 355158878;

  @override
  String toString() => rprint('Literal<$T>($value)');
}

class Variable extends Expression {
  const Variable(this.name);

  final String name;

  dynamic call(Context context) => context.get(name);

  @override
  bool operator ==(dynamic other) => other is Variable && other.name == name;

  @override
  int get hashCode => name.hashCode ^ 355158878;

  @override
  String toString() => 'Variable($name)';
}

class Key extends Expression {
  const Key(this.key, this.expr);

  final Expression key;
  final Expression expr;

  dynamic call(Context context) => expr(context)[key(context)];

  @override
  bool operator ==(dynamic other) =>
      other is Key && other.key == key && other.expr == expr;

  @override
  int get hashCode => key.hashCode ^ expr.hashCode ^ 546874638;

  @override
  String toString() => 'Key($key, $expr)';
}

class Field extends Expression {
  const Field(this.field, this.expr);

  final String field;
  final Expression expr;

  dynamic call(Context context) => getAttr(field, expr(context));

  bool operator ==(dynamic other) =>
      other is Field && other.field == field && other.expr == expr;

  @override
  int get hashCode => field.hashCode ^ expr.hashCode ^ 975090185;

  @override
  String toString() => 'Field($field, $expr)';
}

class Not extends Expression {
  const Not(this.expr);

  final Expression expr;

  bool call(Context context) => !asBool(expr(context));

  bool operator ==(dynamic other) => other is Not && other.expr == expr;

  @override
  int get hashCode => expr.hashCode ^ 176462831;

  @override
  String toString() => 'Not($expr)';
}

class Neg extends Expression {
  const Neg(this.expr);

  final Expression expr;

  dynamic call(Context context) => -expr(context);

  bool operator ==(dynamic other) => other is Neg && other.expr == expr;

  @override
  int get hashCode => expr.hashCode ^ 865004522;

  @override
  String toString() => 'Neg($expr)';
}

class Pos extends Expression {
  const Pos(this.expr);

  final Expression expr;

  dynamic call(Context context) => expr(context);

  bool operator ==(dynamic other) => other is Pos && other.expr == expr;

  @override
  int get hashCode => expr.hashCode ^ 914696886;

  @override
  String toString() => 'Pos($expr)';
}

class Mul extends Expression {
  const Mul(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => left(context) * right(context);

  bool operator ==(dynamic other) =>
      other is Mul && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 544149599;

  @override
  String toString() => 'Mul($left, $right)';
}

class Div extends Expression {
  const Div(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => left(context) / right(context);

  bool operator ==(dynamic other) =>
      other is Div && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 917087792;

  @override
  String toString() => 'Div($left, $right)';
}

class FloorDiv extends Expression {
  const FloorDiv(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => left(context) ~/ right(context);

  bool operator ==(dynamic other) =>
      other is FloorDiv && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 1070751238;

  @override
  String toString() => 'FloorDiv($left, $right)';
}

class Mod extends Expression {
  const Mod(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => left(context) % right(context);

  bool operator ==(dynamic other) =>
      other is Mod && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 677170890;

  @override
  String toString() => 'Mod($left, $right)';
}

class Pow extends Expression {
  const Pow(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) =>
      math.pow(left(context) as num, right(context) as num);

  bool operator ==(dynamic other) =>
      other is Pow && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 1066056896;

  @override
  String toString() => 'Pow($left, $right)';
}

class Add extends Expression {
  const Add(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => left(context) + right(context);

  bool operator ==(dynamic other) =>
      other is Add && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 160579538;

  @override
  String toString() => 'Add($left, $right)';
}

class Sub extends Expression {
  const Sub(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => left(context) - right(context);

  bool operator ==(dynamic other) =>
      other is Add && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 619279262;

  @override
  String toString() => 'Sub($left, $right)';
}

class GreaterOrEqual extends Expression {
  const GreaterOrEqual(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => ge(left(context), right(context));

  bool operator ==(dynamic other) =>
      other is GreaterOrEqual && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 550386602;

  @override
  String toString() => 'GreaterOrEqual($left, $right)';
}

class Greater extends Expression {
  const Greater(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => gt(left(context), right(context));

  bool operator ==(dynamic other) =>
      other is Greater && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 645026259;

  @override
  String toString() => 'Greater($left, $right)';
}

class Less extends Expression {
  const Less(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => lt(left(context), right(context));

  bool operator ==(dynamic other) =>
      other is Less && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 435777655;

  @override
  String toString() => 'Less($left, $right)';
}

class LessOrEqual extends Expression {
  const LessOrEqual(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => le(left(context), right(context));

  bool operator ==(dynamic other) =>
      other is LessOrEqual && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 526628924;

  @override
  String toString() => 'LessOrEqual($left, $right)';
}

class Equal extends Expression {
  const Equal(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => eq(left(context), right(context));

  bool operator ==(dynamic other) =>
      other is Equal && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 697477197;

  @override
  String toString() => 'Equal($left, $right)';
}

class NotEqual extends Expression {
  const NotEqual(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) => ne(left(context), right(context));

  bool operator ==(dynamic other) =>
      other is NotEqual && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 343768363;

  @override
  String toString() => 'NotEqual($left, $right)';
}

class And extends Expression {
  const And(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) {
    dynamic leftRes = left(context);
    return asBool(leftRes) ? right(context) : leftRes;
  }

  bool operator ==(dynamic other) =>
      other is And && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 766773748;

  @override
  String toString() => 'And($left, $right)';
}

class Or extends Expression {
  const Or(this.left, this.right);

  final Expression left;
  final Expression right;

  dynamic call(Context context) {
    dynamic leftRes = left(context);
    return asBool(leftRes) ? leftRes : right(context);
  }

  bool operator ==(dynamic other) =>
      other is Or && other.left == left && other.right == right;

  @override
  int get hashCode => left.hashCode ^ right.hashCode ^ 617640008;

  @override
  String toString() => 'Or($left, $right)';
}

// class Conditional extends Expression {
//   const Conditional(this.expr, this.condition, this.orElse);

//   final Expression expr;
//   final Expression condition;
//   final Expression orElse;

//   dynamic call(Context context) =>
//       asBool(condition(context)) ? expr(context) : orElse(context);

//   bool operator ==(dynamic other) =>
//       other is Conditional &&
//       other.expr == expr &&
//       other.condition == condition &&
//       other.orElse == orElse;

//   @override
//   int get hashCode =>
//       expr.hashCode ^ condition.hashCode ^ orElse.hashCode ^ 749044055;

//   @override
//   String toString() => rprint('Conditional($expr, $condition, $orElse)';
// }

class Call extends Expression {
  const Call(this.expr,
      {this.args = const <Expression>[],
      this.kwargs = const <String, Expression>{}});

  final Expression expr;
  final List<Expression> args;
  final Map<String, Expression> kwargs;

  dynamic call(Context context) => Function.apply(
      expr(context) as Function,
      args.map<dynamic>((arg) => arg(context)).toList(),
      kwargs.map<Symbol, dynamic>((key, expr) =>
          MapEntry<Symbol, dynamic>(Symbol(key), expr(context))));

  bool operator ==(dynamic other) =>
      other is Call &&
      other.expr == expr &&
      other.args == args &&
      other.kwargs == kwargs;

  @override
  int get hashCode =>
      expr.hashCode ^ args.hashCode ^ kwargs.hashCode ^ 799447056;

  @override
  String toString() => 'Call($expr'
      '${args.isNotEmpty ? ', args: ' + args.toString() : ''}'
      '${kwargs.isNotEmpty ? ', kwargs: ' + kwargs.toString() : ''})';
}

class Filter extends Expression {
  const Filter._(this.name, this.expr,
      {this.args = const <Expression>[],
      this.kwargs = const <String, Expression>{}});
  Filter(Environment env, this.name, this.expr,
      {this.args = const <Expression>[],
      this.kwargs = const <String, Expression>{}});

  final String name;
  final Expression expr;
  final List<Expression> args;
  final Map<String, Expression> kwargs;

  @override
  dynamic call(Context context) => context.env.callFilter(
      name,
      <dynamic>[expr(context)]
          .followedBy(args.map<dynamic>((arg) => arg(context)))
          .toList(),
      kwargs.map<Symbol, dynamic>((key, expr) =>
          MapEntry<Symbol, dynamic>(Symbol(key), expr(context))));

  bool operator ==(dynamic other) =>
      other is Filter &&
      other.name == name &&
      other.expr == expr &&
      other.args == args &&
      other.kwargs == kwargs;

  @override
  int get hashCode =>
      name.hashCode ^
      expr.hashCode ^
      args.hashCode ^
      kwargs.hashCode ^
      430908456;

  String toString() => 'Filter($name, $expr'
      '${args.isNotEmpty ? ', args: ' + args.toString() : ''}'
      '${kwargs.isNotEmpty ? ', kwargs: ' + kwargs.toString() : ''})';
}

class Test extends Expression {
  const Test._(this.name, this.expr,
      {this.args = const <Expression>[],
      this.kwargs = const <String, Expression>{}});
  Test(Environment env, this.name, this.expr,
      {this.args = const <Expression>[],
      this.kwargs = const <String, Expression>{}});

  final String name;
  final Expression expr;
  final List<Expression> args;
  final Map<String, Expression> kwargs;

  @override
  bool call(Context context) => context.env.callTest(
      name,
      <dynamic>[expr(context)]
          .followedBy(args.map<dynamic>((arg) => arg(context)))
          .toList(),
      kwargs.map<Symbol, dynamic>((key, expr) =>
          MapEntry<Symbol, dynamic>(Symbol(key), expr(context)))) as bool;

  bool operator ==(dynamic other) =>
      other is Test &&
      other.name == name &&
      other.expr == expr &&
      other.args == args &&
      other.kwargs == kwargs;

  @override
  int get hashCode =>
      name.hashCode ^
      expr.hashCode ^
      args.hashCode ^
      kwargs.hashCode ^
      648260012;

  @override
  String toString() => 'Test($name, $expr'
      '${args.isNotEmpty ? ', args: ' + args.toString() : ''}'
      '${kwargs.isNotEmpty ? ', kwargs' + kwargs.toString() : ''})';
}

// Statements

class Extend extends Node {
  Extend(this.env, this.parent, this.blocks);

  final Environment env;
  final Expression parent;
  final Map<String, Block> blocks;

  @override
  String render(Context context) {
    Context extendContext = context.copyWith(blocks: blocks);
    String parentPath = parent(extendContext);
    return env.callTemplateWithContext(parentPath, extendContext);
  }

  bool operator ==(dynamic other) =>
      other is Extend && other.parent == parent && other.blocks == blocks;

  @override
  int get hashCode => parent.hashCode ^ blocks.hashCode ^ 1038140602;

  @override
  String toString() => 'Extend($parent, $blocks)';
}

class Block extends Node {
  const Block(this.name, this.body);

  final String name;
  final Node body;

  @override
  String render(Context context) {
    if (context.blocks.containsKey(name)) {
      Context blockContext = context.copyWith(
          data: Map<String, dynamic>.of(context.data)
            ..addAll(<String, dynamic>{'super': () => body.render(context)}));
      return context.blocks[name].body.render(blockContext);
    }
    return body.render(context);
  }

  bool operator ==(dynamic other) =>
      other is Block && other.name == name && other.body == body;

  @override
  int get hashCode => name.hashCode ^ body.hashCode ^ 62911841;

  @override
  String toString() => 'Block($name, $body)';
}

class For extends Node {
  const For(this.target, this.iter, this.body,
      {this.isRecursive, this.test, this.orElse});

  final String target;
  final Expression iter;
  final Node body;
  final bool isRecursive;
  final Test test;
  final Node orElse;

  @override
  String render(Context context) {
    dynamic iter = this.iter(context);

    if (iter is Iterable) if (iter.isNotEmpty) {
      List<dynamic> list = iter.toList();
      StringBuffer buffer = StringBuffer();

      dynamic prevItem;
      bool changed(dynamic newitem) {
        bool isChanged = prevItem != newitem;
        prevItem = newitem;
        return isChanged;
      }

      LoopContext loopContext;
      Context ctx;

      if (test != null)
        for (int i = 0; i < list.length; i++) {
          loopContext = LoopContext(
              list.length,
              i,
              i > 0 ? list[i - 1] : const Undefined(),
              i < list.length - 1 ? list[i + 1] : const Undefined(),
              changed);
          ctx = context.copyWith(
              data: Map<String, dynamic>.of(context.data)
                ..addAll(
                    <String, dynamic>{target: list[i], 'loop': loopContext}));
          if (test(ctx)) buffer.write(body.render(ctx));
        }
      else
        for (int i = 0; i < list.length; i++) {
          loopContext = LoopContext(
              list.length,
              i,
              i > 0 ? list[i - 1] : const Undefined(),
              i < list.length - 1 ? list[i + 1] : const Undefined(),
              changed);
          ctx = context.copyWith(
              data: Map<String, dynamic>.of(context.data)
                ..addAll(
                    <String, dynamic>{target: list[i], 'loop': loopContext}));
          buffer.write(body.render(ctx));
        }
      return buffer.toString();
    }

    return orElse != null ? orElse.render(context) : throw ArgumentError();
  }

  bool operator ==(dynamic other) =>
      other is For &&
      other.target == target &&
      other.iter == iter &&
      other.isRecursive == isRecursive &&
      other.body == body &&
      other.test == test &&
      other.orElse == orElse;

  @override
  int get hashCode =>
      target.hashCode ^
      iter.hashCode ^
      body.hashCode ^
      isRecursive.hashCode ^
      test.hashCode ^
      orElse.hashCode ^
      589583422;

  @override
  String toString() => 'For($target, $iter, $body'
      '${isRecursive ? ', isRecursive: ' + isRecursive.toString() : ''}'
      '${test != null ? ', test: ' + test.toString() : ''}'
      '${orElse != null ? ', orElse: ' + orElse.toString() : ''})';
}

class If extends Node {
  const If(this.test, this.body, [this.orElse]);

  final Test test;
  final Node body;
  final Node orElse;

  @override
  String render(Context context) => test(context)
      ? body.render(context)
      : orElse != null ? orElse.render(context) : '';

  bool operator ==(dynamic other) =>
      other is If &&
      other.test == test &&
      other.body == body &&
      other.orElse == orElse;

  @override
  int get hashCode =>
      test.hashCode ^ body.hashCode ^ orElse.hashCode ^ 1049012907;

  @override
  String toString() => 'If($test, $body'
      '${orElse != null ? ', ' + orElse.toString() : ''})';
}

// Helpers

class Tag extends Node {
  const Tag(this.tag, [this.arg, this.arg2, this.arg3, this.arg4]);

  final String tag;
  final dynamic arg;
  final dynamic arg2;
  final dynamic arg3;
  final dynamic arg4;

  @override
  bool operator ==(dynamic other) =>
      other is Tag &&
      other.tag == tag &&
      other.arg == arg &&
      other.arg2 == arg2 &&
      other.arg3 == arg3 &&
      other.arg4 == arg4;

  @override
  int get hashCode =>
      tag.hashCode ^
      arg.hashCode ^
      arg2.hashCode ^
      arg3.hashCode ^
      arg4.hashCode ^
      67728975;

  @override
  String toString() => 'Tag($tag'
      '${arg != null ? ', arg: ' + arg.toString() : ''}'
      '${arg2 != null ? ', arg2: ' + arg2.toString() : ''}'
      '${arg3 != null ? ', arg3: ' + arg3.toString() : ''}'
      '${arg4 != null ? ', arg4: ' + arg4.toString() : ''})';
}
