import 'package:jinja/jinja.dart';
import "package:test/test.dart";

void main() {
  group('For Loop', () {
    final Environment env = Environment();

    test("Simple", () {
      Template template =
          env.fromSource('{% for item in seq %}{{ item }}{% endfor %}');

      expect(
          template.render(<String, dynamic>{
            'seq': List<int>.generate(10, (i) => i),
          }),
          equals('0123456789'));
    });

    test("Else", () {
      Template template =
          env.fromSource('{% for item in seq %}XXX{% else %}...{% endfor %}');

      expect(template.render(), equals('...'));
    });

    test("Else Scoping Item", () {
      Template template = env
          .fromSource('{% for item in [] %}{% else %}{{ item }}{% endfor %}');

      expect(template.render(<String, dynamic>{'item': 42}), equals('42'));
    });

    test("Empty Blocks", () {
      Template template =
          env.fromSource('<{% for item in seq %}{% else %}{% endfor %}>');

      expect(template.render(), equals('<>'));
    });

    test("Context Vars", () {
      List<int> slist = [42, 24];

      Template template;
      for (Object seq in [slist, slist.reversed]) {
        template = env.fromSource('{% for item in seq %}{{ loop.index }}|'
            '{{ loop.index0 }}|{{ loop.revindex }}|{{loop.revindex0 }}|'
            '{{ loop.first }}|{{ loop.last }}|{{ loop.length }}###'
            '{% endfor %}');

        List<String> parts =
            template.render(<String, dynamic>{'seq': seq}).split('###');
        List<String> one = parts[0].split('|');
        List<String> two = parts[1].split('|');

        expect(one[0], equals('1'));
        expect(one[1], equals('0'));
        expect(one[2], equals('2'));
        expect(one[3], equals('1'));
        expect(one[4], equals('true'));
        expect(one[5], equals('false'));
        expect(one[6], equals('2'));

        expect(two[0], equals('2'));
        expect(two[1], equals('1'));
        expect(two[2], equals('1'));
        expect(two[3], equals('0'));
        expect(two[4], equals('false'));
        expect(two[5], equals('true'));
        expect(two[6], equals('2'));
      }
    });

    test("Cycling", () {
      Template template = env.fromSource('{% for item in seq %}'
          '{{ loop.cycle("<1>", "<2>") }}{% endfor %}{% for item in seq %}'
          '{{ loop.cycle(*through) }}{% endfor %}');

      expect(
          template.render(<String, dynamic>{
            'seq': List<int>.generate(4, (int i) => i),
            'through': <String>['<1>', '<2>']
          }),
          equals('<1><2>' * 4));
    });

    test("LookAround", () {
      Template template = env.fromSource('{% for item in seq %}'
          '{{ loop.previtem | default("x") }}-{{ item }}-{{'
          'loop.nextitem | default("x") }}|{% endfor %}');

      expect(
          template.render(
              <String, dynamic>{'seq': List<int>.generate(4, (int i) => i)}),
          equals('x-0-1|0-1-2|1-2-3|2-3-x|'));
    });

    test("Changed", () {
      Template template = env.fromSource('{% for item in seq %}'
          '{{ loop.changed(item) }},{% endfor %}');

      expect(
          template.render(<String, dynamic>{
            'seq': <dynamic>[null, null, 1, 2, 2, 3, 4, 4, 4]
          }),
          equals('true,false,true,true,false,true,true,false,false,'));
    });

    test("Scope", () {
      Template template =
          env.fromSource('{% for item in seq %}{% endfor %}{{ item }}');

      expect(
          template.render(
              <String, dynamic>{'seq': List<int>.generate(10, (int i) => i)}),
          equals(''));
    });

    test("VarLen", () {
      Template template =
          env.fromSource('{% for item in iter %}{{ item }}{% endfor %}');

      Iterable<int> inner() sync* {
        for (int i = 0; i < 5; i++) yield i;
      }

      expect(
          template.render(<String, dynamic>{'iter': inner()}), equals('01234'));
    });

    test("NonIter", () {
      Template template =
          env.fromSource('{% for item in none %}...{% endfor %}');

      expect(() => template.render(), throwsArgumentError);
    });

    test("Recursive", () {
      Template template = env.fromSource('{% for item in seq recursive %}'
          '[{{ item["a"] }}{% if item["b"] %}<{{ loop(item["b"]) }}>{% endif %}]'
          '{% endfor %}');

      expect(
          template.render(<String, dynamic>{
            'seq': [
              <String, dynamic>{
                'a': 1,
                'b': [
                  <String, dynamic>{'a': 1},
                  <String, dynamic>{'a': 2},
                ]
              },
              <String, dynamic>{
                'a': 2,
                'b': [
                  <String, dynamic>{'a': 1},
                  <String, dynamic>{'a': 2},
                ],
              },
              <String, dynamic>{
                'a': 3,
                'b': [
                  <String, dynamic>{'a': 'a'},
                ],
              },
            ],
          }),
          equals('[1<[1][2]>][2<[1][2]>][3<[a]>]'));
    });
  });

  group('If Condition', () {
    final Environment env = Environment();

    test("Simple", () {
      Template template = env.fromSource('{% if true %}...{% endif %}');

      expect(template.render(), equals('...'));
    });

    test("Elif", () {
      Template template = env.fromSource(
          '{% if false %}XXX{% elif true %}...{% else %}XXX{% endif %}');

      expect(template.render(), equals('...'));
    });

    test("Elif Deep", () {
      String source = '{% if a == 0 %}0' +
          List<String>.generate(1000, (i) => '{% elif a == $i %}$i').join() +
          '{% else %}x{% endif %}';
      Template template = env.fromSource(source);

      expect(template.render(<String, dynamic>{'a': 0}), equals('0'));
      expect(template.render(<String, dynamic>{'a': 10}), equals('10'));
      expect(template.render(<String, dynamic>{'a': 999}), equals('999'));
      expect(template.render(<String, dynamic>{'a': 1000}), equals('x'));
    });

    test("Else", () {
      Template template =
          env.fromSource('{% if false %}XXX{% else %}...{% endif %}');

      expect(template.render(), equals('...'));
    });

    test("Empty", () {
      Template template =
          env.fromSource('[{% if true %}{% else %}{% endif %}]');

      expect(template.render(), equals('[]'));
    });

    test("Complete", () {
      Template template = env.fromSource(
          '{% if a %}A{% elif b %}B{% elif c == d %}C{% else %}D{% endif %}');

      expect(
          template.render(
              <String, dynamic>{'a': 0, 'b': false, 'c': 42, 'd': 42.0}),
          equals('C'));
    });

    test("No Scope", () {
      Template template =
          env.fromSource('{% if a %}{% set foo = 1 %}{% endif %}{{ foo }}');

      expect(template.render(<String, dynamic>{'a': true}), equals('1'));
      template =
          env.fromSource('{% if true %}{% set foo = 1 %}{% endif %}{{ foo }}');
      expect(template.render(), equals('1'));
    });
  });
}
